<?php

namespace FiloBlu\Refilo\Remote\Entity;

use Exception;
use FiloBlu\Refilo\Helper\Catalog\ProductAttributeHelper;
use FiloBlu\Refilo\Helper\Catalog\ProductHelper;
use Magento\Catalog\Api\Data\ProductInterface;
use Magento\Catalog\Model\Product\Attribute\Source\Status as AttributeStatus;
use Magento\Catalog\Model\Product\Option;
use Magento\Catalog\Model\Product\Type;
use Magento\Catalog\Model\Product\Visibility;
use Magento\ConfigurableProduct\Api\Data\OptionInterface;
use Magento\ConfigurableProduct\Model\Product\Type\Configurable;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\GiftCard\Model\Catalog\Product\Type\Giftcard;
use Magento\GroupedProduct\Model\Product\Type\Grouped;

/**
 * Class MagentoToRefiloProductAdapter
 * @package FiloBlu\Refilo\Remote\Entity
 */
class MagentoToRefiloProductAdapter
{
    /**
     * @var ProductAttributeHelper
     */
    private $productAttributeHelper;

    /**
     * @var ProductHelper
     */
    private $productHelper;

    /**
     * @var Option
     */
    private $option;

    /**
     * MagentoToRefiloProductAdapter constructor.
     * @param ProductAttributeHelper $productAttributeHelper
     * @param ProductHelper $productHelper
     * @param Option $option
     */
    public function __construct(
        ProductAttributeHelper $productAttributeHelper,
        ProductHelper $productHelper,
        Option $option
    )
    {
        $this->productAttributeHelper = $productAttributeHelper;
        $this->productHelper = $productHelper;
        $this->option = $option;
    }

    /**
     * @param \Magento\Catalog\Model\Product|ProductInterface $magentoProduct
     * @param Product $refiloProduct
     * @return Product
     * @throws NoSuchEntityException
     * @throws Exception
     */
    public function adapt($magentoProduct, Product $refiloProduct): Product
    {
        $docType = Product::DOCTYPE_SPEC_DISABLED;
        $visibility = '0';
        $searchVisibility = '0';

        if ((int)$magentoProduct->getStatus() === AttributeStatus::STATUS_ENABLED) {
            $docType = Product::DOCTYPE_SPEC_ENABLED;
            $visibility = (!$this->isProductVisible($magentoProduct) ? '0' : '1');
            $searchVisibility = (!$this->isProductVisibleInSearch($magentoProduct) ? '0' : '1');
        }

        $data = [
            Product::DOCTYPE           => $docType,
            Product::ID                => (int)$magentoProduct->getEntityId(),
            Product::ENABLED           => $this->isEnabled($magentoProduct),
            Product::SKU               => $magentoProduct->getSku(),
            Product::TYPE              => $magentoProduct->getTypeId(),
            Product::NAME              => $magentoProduct->getName(),
            Product::CREATED_AT        => $magentoProduct->getCreatedAt(),
            Product::UPDATED_AT        => $magentoProduct->getUpdatedAt(),
            Product::WEIGHT            => $magentoProduct->getWeight(),
            Product::META_TITLE        => $magentoProduct->getMetaTitle(),
            Product::META_KEYWORDS     => $magentoProduct->getMetaKeywords(),
            Product::META_DESCRIPTION  => $magentoProduct->getMetaDescription(),
            Product::VISIBLE           => $visibility,
            Product::ATTRIBUTES        => $this->productAttributeHelper->getAttributesValues($magentoProduct),
            Product::FILTERS           => $this->productAttributeHelper->getFilterableAttributes($magentoProduct),
            Product::SORTABLES         => $this->productAttributeHelper->getSortableAttributes($magentoProduct),
            Product::UPSELL            => $magentoProduct->getUpSellProductIds(),
            Product::CROSSSELL         => $magentoProduct->getCrossSellProductIds(),
            Product::RELATED           => $magentoProduct->getRelatedProductIds(),
            Product::ASSETS            => $this->getAssets($magentoProduct),
            Product::INVENTORY         => $this->productHelper->getStockFor($magentoProduct),
            Product::SEARCH_VISIBILITY => $searchVisibility,
            Product::CUSTOM_OPTIONS    => $this->getCustomOptions($magentoProduct),
            Product::CATEGORIES        => [],
            Product::CATEGORY_DATA     => []
        ];

        $refiloProduct->setData(
            array_merge(
                $data,
                $this->productHelper->getPricesFor($magentoProduct),
                $this->productHelper->getCategoriesFor($magentoProduct)
            )
        );

        return $this->fillChildren($magentoProduct, $refiloProduct);
    }

    /**
     * @param ProductInterface|\Magento\Catalog\Model\Product $magentoProduct
     * @return bool
     */
    public function isProductVisible($magentoProduct): bool
    {
        switch ((int)$magentoProduct->getVisibility()) {
            case Visibility::VISIBILITY_IN_CATALOG:
            case Visibility::VISIBILITY_BOTH:
                return true;
        }

        return false;
    }

    /**
     * @param ProductInterface|\Magento\Catalog\Model\Product $magentoProduct
     * @return bool
     */
    public function isProductVisibleInSearch($magentoProduct): bool
    {
        switch ((int)$magentoProduct->getVisibility()) {
            case Visibility::VISIBILITY_IN_SEARCH:
            case Visibility::VISIBILITY_BOTH:
                return true;
        }

        return false;
    }

    /**
     * @param \Magento\Catalog\Model\Product|ProductInterface $magentoProduct
     * @return string
     */
    public function isEnabled($magentoProduct): string
    {
        return ((int)$magentoProduct->getStatus() === AttributeStatus::STATUS_ENABLED) ? '1' : '0';
    }

    /**
     * @param \Magento\Catalog\Model\Product|ProductInterface $magentoProduct
     * @return array
     */
    protected function getAssets($magentoProduct): array
    {
        $assets = [];
        $gallery = $magentoProduct->getData('media_gallery');

        foreach ($gallery['images'] as $image) {
            // Skip image is hide
            if (isset($image['disabled']) && $image['disabled'] == 1) {
                continue;
            }
            $assets['images'][] = [
                'label'             => $image['label'] ?? null, // TODO : remove?
                'alt'               => $image['label'] ?? null,
                'file'              => ($image['file']) ? "/catalog/product{$image['file']}" : null,
                'media_type'        => $image['media_type'] ?? null,
                'video_url'         => $image['video_url'] ?? null,
                'video_title'       => $image['video_title'] ?? null,
                'video_description' => $image['video_description'] ?? null,
                'video_metadata'    => $image['video_metadata'] ?? null,
                'video_provider'    => $image['video_provider'] ?? null,
            ];
        }

        $assets['placeholders'] = $this->productAttributeHelper->getImagePlaceholdersAttributes($magentoProduct);
        return $assets;
    }

    /**
     * @param \Magento\Catalog\Model\Product $magentoProduct
     * @return array
     */
    public function getCustomOptions(\Magento\Catalog\Model\Product $magentoProduct): array
    {
        $customOptions = [];

        // TODO: memory leaked
        //
        // Do not call following collection because memory leak
        // $collection = $magentoProduct->getProductOptionsCollection();

        /** @var \Magento\Catalog\Model\ResourceModel\Product\Option $item */
        foreach ($this->option->getProductOptionCollection($magentoProduct)->getItems() as $item) {
            $data = $item->getData();
            $data['values'] = [];

            foreach ($item->getValues() ?? [] as $value) {
                $data['values'][] = $value->getData();
            }

            $customOptions[] = $data;
        }
        return $customOptions;
    }

    /**
     * @param \Magento\Catalog\Model\Product $magentoProduct
     * @param Product $refiloProduct
     * @return Product
     * @throws Exception
     */
    protected function fillChildren(\Magento\Catalog\Model\Product $magentoProduct, Product $refiloProduct): Product
    {
        switch ($magentoProduct->getTypeId()) {
            case Configurable::TYPE_CODE:
                return $this->getConfigurableChildren($magentoProduct, $refiloProduct);
            case Type::TYPE_BUNDLE:
                return $this->getBundleChildren($magentoProduct, $refiloProduct);
            case Grouped::TYPE_CODE:
                return $this->getGroupedChildren($magentoProduct, $refiloProduct);
            case Giftcard::TYPE_GIFTCARD:
                return $this->getGiftCardsChildren($magentoProduct, $refiloProduct);
        }
        return $refiloProduct;
    }

    /**
     * @param \Magento\Catalog\Model\Product $magentoProduct
     * @param Product $refiloProduct
     * @return Product
     * @throws Exception
     */
    public function getConfigurableChildren(\Magento\Catalog\Model\Product $magentoProduct, Product $refiloProduct): Product
    {
        if ($magentoProduct->getTypeId() !== Configurable::TYPE_CODE) {
            return $refiloProduct;
        }

        $chidrenIds = [];
        $sellableAttributes = [];
        $storeId = $magentoProduct->getStore()->getId();
        $extensionAttributes = $magentoProduct->getExtensionAttributes();


        // TODO: memory leaked
        //
        // $configurable = $magentoProduct->getTypeInstance();
        // $usedAttributes = $configurable->getUsedProductAttributes($magentoProduct);
        //
        $usedAttributes = $this->productHelper->getUsedProductAttributes($magentoProduct);


        // TODO: memory leaked
        //
        // $configurable = $magentoProduct->getTypeInstance();
        // $productsChildren = $configurable->getUsedProducts($magentoProduct);
        //
        $productsChildren = $this->productHelper->getUsedProducts($magentoProduct, $usedAttributes);



        foreach ($usedAttributes as $attribute) {
            $sellableAttributes[$attribute['attribute_code']] = [];
        }

        foreach ($productsChildren as $productChild) {
            if ((int)$productChild->getStatus() !== AttributeStatus::STATUS_ENABLED) {
                continue;
            }

            foreach ($sellableAttributes as $attributeCode => $data) {
                $sellableAttributes[$attributeCode][$productChild->getData($attributeCode)] = true;
            }

            $chidrenIds[] = (int)$productChild->getId();
        }

        unset($productChild);

        $refiloProduct->setChildren($chidrenIds);
        $productOptions = $extensionAttributes->getConfigurableProductOptions() ?? [];

        $variants = [];
        foreach ($productOptions as $option) {
            $attributeCode = $option->getProductAttribute()->getAttributeCode();
            $variant = [
                'id'             => $option->getId(),
                'type'           => $option->getProductAttribute()->getData('frontend_input'),
                'position'       => $option->getPosition(),
                'label'          => $option->getProductAttribute()->getStoreLabel($storeId),
                'use_default'    => (int)$option->getIsUseDefault(),
                'attribute_id'   => $option->getAttributeId(),
                'attribute_code' => $attributeCode,
                'values'         => $this->getOptions(
                    $option,
                    $sellableAttributes[$attributeCode],
                    $storeId,
                    $attributeCode
                ),
                'options'        => []
            ];

            // TODO: use magento deserializer instead
            $additionalData = json_decode(
                $option->getProductAttribute()->getData('additional_data'),
                true
            );

            if ($additionalData) {
                $variant['rendered'] = $additionalData['swatch_input_type'] ?? '';
            }

            $options = [];
            foreach ($variant['values'] as $value) {
                $options[] = (int)$value['value_index'];
            }

            $variant['options'] = $options;
            $variants[] = $variant;
        }

        $filters = [];
        foreach ($productsChildren as $salableChild) {

            /** @var ProductInterface $salableChild */
            if (!$salableChild->isSalable()){
                continue;
            }
            $filterable = $this->productAttributeHelper->getFilterableAttributes($salableChild);

            foreach ($filterable as $attributeCode => $value) {
                // TODO: handle cases which not are multiselect or select
                if (!isset($value['key']) || !$value['key']) {
                    continue;
                }

                if (!isset($filters[$attributeCode])) {
                    $filters[$attributeCode] = $value;
                    unset($filters[$attributeCode]['swatch']);

                    if (!is_array($value['value'])) {
                        $filters[$attributeCode]['value'] = [$value['value']];
                    }

                    if (!is_array($value['key'])) {
                        $filters[$attributeCode]['key'] = [$value['key']];
                    }

                    continue;
                }

                if (is_array($value['value'])) {
                    $filters[$attributeCode]['value'] =
                        array_merge($filters[$attributeCode]['value'], $value['value']);
                    $filters[$attributeCode]['key'] =
                        array_merge($filters[$attributeCode]['key'], $value['key']);
                    continue;
                }

                $filters[$attributeCode]['value'][] = $value['value'];
                $filters[$attributeCode]['key'][] = $value['key'];
            }
        }

        $parentFilters = $refiloProduct->getData(Product::FILTERS);
        $refiloProduct->setData(Product::FILTERS, array_merge($parentFilters, $filters));
        $refiloProduct->setVariants($variants);
        $productOptions = null;
        return $refiloProduct;
    }

    /**
     * @param OptionInterface $option
     * @param $usableAttributes
     * @param $storeId
     * @param $attributeCode
     * @return array
     */
    protected function getOptions(OptionInterface $option, $usableAttributes, $storeId, $attributeCode)
    {
        $output = [];

        foreach ($option->getOptions() ?? [] as $optionValue) {
            if (!array_key_exists($optionValue['value_index'], $usableAttributes)) {
                continue;
            }
            $data = $this->productAttributeHelper->getSwatchByOptionId($optionValue['value_index']);
            $data['value_index'] = $optionValue['value_index'];
            $data['label'] = $this->productAttributeHelper->getProductOptionLabel($optionValue['value_index'], $storeId, $attributeCode);
            $output[] = $data;
        }

        return $output;
    }

    /**
     * @param ProductInterface $magentoProduct
     * @param Product $refiloProduct
     * @return Product
     */
    public function getBundleChildren(ProductInterface $magentoProduct, Product $refiloProduct)
    {
        $extensionAttributes = $magentoProduct->getExtensionAttributes();
        $options = [];

        $refiloProduct->setData('bundle_price_type', (int)$magentoProduct->getPriceType() == 0 ? 'dynamic' : 'fixed');
        foreach ($extensionAttributes->getBundleProductOptions() as $bundleProductOption) {
            $productLinks = $bundleProductOption->getData('product_links');

            if (empty($productLinks)) {
                continue;
            }

            $option = [
                'id'       => (int)$bundleProductOption->getData('option_id'),
                'title'    => $bundleProductOption->getData('title'),
                'required' => (int)$bundleProductOption->getData('required'),
                'type'     => $bundleProductOption->getData('type'),
                'position' => (int)$bundleProductOption->getData('position'),
                'childs'   => []
            ];

            foreach ($productLinks as $productLink) {
                $link_data = $productLink->getData();
                $link_data['product_id'] = $link_data['entity_id'];
                $link_data['product_sku'] = $link_data['sku'];

                unset($link_data['entity_id']);
                unset($link_data['sku']);
                unset($link_data['option_id']);

                $child = [];
                foreach ($link_data as $key => $value) {
                    $child[$key] = (is_numeric($value)) ? $value + 0 : $value;
                }

                $option['childs'][] = $child;
            }

            $options[] = $option;
        }

        $refiloProduct->setBundle($options);
        return $refiloProduct;
    }

    /**
     * @param ProductInterface $magentoProduct
     * @param Product $refiloProduct
     * @return Product
     */
    public function getGroupedChildren(ProductInterface $magentoProduct, Product $refiloProduct): Product
    {
        $links = $magentoProduct->getTypeInstance()->getChildrenIds($magentoProduct->getId());

        if (empty($links)) {
            return $refiloProduct;
        }

        $refiloProduct->setChildren(array_values(array_shift($links)));
        return $refiloProduct;
    }

    /**
     * @param ProductInterface|\Magento\Catalog\Model\Product $magentoProduct
     * @param Product $refiloProduct
     * @return Product
     */
    public function getGiftCardsChildren($magentoProduct, Product $refiloProduct): Product
    {
        $giftCardsAmounts = [];
        $hasWebsiteValue = false;

        foreach ($magentoProduct->getGiftcardAmounts() as $amount) {

            if (!$hasWebsiteValue && (int)$amount['website_id'] > 0) {
                $hasWebsiteValue = true;
                break;
            }
        }

        foreach ($magentoProduct->getGiftcardAmounts() as $data) {
            if ($hasWebsiteValue && (int)$data['website_id'] === 0) {
                continue;
            }

            $giftCardsAmounts[] = $data;
        }

        $attributes = $refiloProduct->getData('attributes');
        $attributes['giftcard_amounts'] = [
            'title' => 'Amount',
            'value' => $giftCardsAmounts
        ];

        $refiloProduct->setData('attributes', $attributes);
        return $refiloProduct;
    }
}
